package org.eclipse.wtp.releng.cvslog.tools;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;
import java.util.StringTokenizer;

import org.eclipse.wtp.releng.cvstools.Delta;
import org.eclipse.wtp.releng.cvstools.RlogDocument;

public class CVSRlog {

	/**
	 * name of the official cvs as returned as part of the 'cvs version' command
	 * output
	 */
	static final String OFFICIAL_CVS_NAME = "CVS";

	/**
	 * Represents the version of a CVS client or server
	 */
	public static class Version {
		private final String cvsName;

		private final String cvsVersion;

		public Version(String name, String version) {
			if (name == null) {
				throw new IllegalArgumentException("name can't be null");
			}
			if (version == null) {
				throw new IllegalArgumentException("version can't be null");
			}
			this.cvsName = name;
			this.cvsVersion = version;
		}

		public String getCvsName() {
			return cvsName;
		}

		public String getCvsVersion() {
			return cvsVersion;
		}

		public boolean equals(Object o) {
			if (this == o) {
				return true;
			} else if (!(o instanceof Version)) {
				return false;
			}

			final Version version = (Version) o;

			if (!cvsName.equals(version.cvsName)) {
				return false;
			} else if (!cvsVersion.equals(version.cvsVersion)) {
				return false;
			}

			return true;
		}

		public int hashCode() {
			int result;
			result = cvsName.hashCode();
			result = 29 * result + cvsVersion.hashCode();
			return result;
		}

		public String toString() {
			return cvsName + " " + cvsVersion;
		}
	}

	private Hashtable properties = new Hashtable();

	private String property;

	private String propertyOnDelete;


	/**
	 * The caller must provide the CVSROOT to use when calling CVS.
	 */
	private String cvsroot;

	/**
	 * The caller must provide the modules to use when calling CVS.
	 */
	private String[] modules;

	/**
	 * The CVS tag we are dealing with.
	 */
	private String tag;

	/**
	 * The version of the cvs server
	 */
	private Version cvsServerVersion;

	/**
	 * This line delimits seperate files in the CVS log information.
	 */
	private static final String CVS_FILE_DELIM = "=============================================================================";

	/**
	 * This is the keyword that precedes the name of the RCS filename in the CVS
	 * log information.
	 */
	private static final String CVS_RCSFILE_LINE = "RCS file: ";

	/**
	 * This is the keyword that precedes the name of the working filename in the
	 * CVS log information.
	 */
	private static final String CVS_WORKINGFILE_LINE = "Working file: ";

	/**
	 * This line delimits the different revisions of a file in the CVS log
	 * information.
	 */
	private static final String CVS_REVISION_DELIM = "----------------------------";

	/**
	 * This is the keyword that precedes the timestamp of a file revision in the
	 * CVS log information.
	 */
	private static final String CVS_REVISION_DATE = "date:";

	/**
	 * This is the name of the tip of the main branch, which needs special
	 * handling with the log entry parser
	 */
	private static final String CVS_HEAD_TAG = "HEAD";

	/**
	 * This is the keyword that tells us when we have reaced the ned of the
	 * header as found in the CVS log information.
	 */
	private static final String CVS_DESCRIPTION = "description:";

	/**
	 * This is the keyword that tells us when we have reaced the end of the
	 * header as found in the CVS log information.
	 */
	private static final String CVS_HEAD = "head:";
	
	/**
	 * This is the keyword that tells us when we have reaced the ned of the
	 * branch as found in the CVS log information.
	 */
	private static final String CVS_BRANCH = "branch:";

	/**
	 * This is a state keyword which indicates that a revision to a file was not
	 * relevant to the current branch, or the revision consisted of a deletion
	 * of the file (removal from branch..).
	 */
	private static final String CVS_REVISION_DEAD = "dead";

	/**
	 * System dependent new line seperator.
	 */
	private static final String NEW_LINE = System.getProperty("line.separator");

	/**
	 * This is the date format required by commands passed to CVS.
	 */
	static final SimpleDateFormat CVSDATE = new SimpleDateFormat(
			"yyyy-MM-dd HH:mm:ss 'GMT'");

	/**
	 * This is the date format returned in the log information from CVS.
	 */
	static final SimpleDateFormat LOGDATE = new SimpleDateFormat(
			"yyyy/MM/dd HH:mm:ss z");

	static {
		// The timezone is hard coded to GMT to prevent problems with it being
		// formatted as GMT+00:00. However, we still need to set the time zone
		// of the formatter so that it knows it's in GMT.
		CVSDATE.setTimeZone(java.util.TimeZone.getTimeZone("GMT"));
	}

	/**
	 * Sets the CVSROOT for all calls to CVS.
	 * 
	 * @param cvsroot
	 *            CVSROOT to use.
	 */
	public void setCvsRoot(String cvsroot) {
		this.cvsroot = cvsroot;
	}

	/**
	 * Set the cvs tag. Note this should work with names, numbers, and anything
	 * else you can put on log -rTAG
	 * 
	 * @param tag
	 *            the cvs tag
	 */
	public void setTag(String tag) {
		this.tag = tag;
	}

	public void setProperty(String property) {
		this.property = property;
	}

	public void setPropertyOnDelete(String propertyOnDelete) {
		this.propertyOnDelete = propertyOnDelete;
	}

	protected Version getCvsServerVersion() {
		if (cvsServerVersion == null) {

			String[] exec = {"cvs","-d",cvsroot,"version"};

			Process p = null;
			try {

				p = Runtime.getRuntime().exec(exec);
				logErrorStream(p);
				InputStream is = p.getInputStream();
				BufferedReader in = new BufferedReader(
						new InputStreamReader(is));

				String line = in.readLine();
				if (line.startsWith("Client:")) {
					line = in.readLine();
					if (!line.startsWith("Server:")) {
						System.err
								.println("Warning expected a line starting with \"Server:\" but got "
										+ line);
						// we try anyway
					}
				}
				Trace.trace(Trace.FINEST, "server version line: " + line);
				int nameBegin = line.indexOf(" (");
				int nameEnd = line.indexOf(") ", nameBegin);
				final String name;
				final String version;
				if (nameBegin == -1 || nameEnd < nameBegin
						|| nameBegin + 2 >= line.length()) {
					name = "";
				} else {
					name = line.substring(nameBegin + 2, nameEnd);
				}
				int verEnd = line.indexOf(" ", nameEnd + 2);
				if (verEnd < nameEnd + 2) {
					version = "";
				} else {
					version = line.substring(nameEnd + 2, verEnd);
				}

				cvsServerVersion = new Version(name, version);

				Trace.trace(Trace.FINEST, "cvs server version: "
						+ cvsServerVersion);

				p.waitFor();
				p.getInputStream().close();
				p.getOutputStream().close();
				p.getErrorStream().close();
			} catch (Exception e) {
				Trace.trace(Trace.SEVERE, "Failed reading cvs server version",
						e);
			}

			if (p == null || p.exitValue() != 0) {
				if (p == null) {
					Trace.trace(Trace.FINEST,
							"Process p was null in CVS.getCvsServerVersion()");
				} else {
					Trace.trace(Trace.FINEST, "Process exit value = "
							+ p.exitValue());
				}
				cvsServerVersion = new Version(OFFICIAL_CVS_NAME, "1.11");
				Trace.trace(Trace.FINEST,
						"problem getting cvs server version; using "
								+ cvsServerVersion);
			}
		}
		return cvsServerVersion;
	}

	public boolean isCvsNewOutputFormat() {
		Version version = getCvsServerVersion();
		if (OFFICIAL_CVS_NAME.equals(version.getCvsName())) {
			String csv = version.getCvsVersion();
			StringTokenizer st = new StringTokenizer(csv, ".");
			try {
				st.nextToken();
				int subversion = Integer.parseInt(st.nextToken());
				if (subversion > 11) {
					if (subversion == 12) {
						if (Integer.parseInt(st.nextToken()) < 9) {
							return false;
						}
					}
					return true;
				}
			} catch (Throwable e) {
				Trace
						.trace(Trace.FINEST,
								"problem identifying cvs server. Assuming output is of 'old' type");
			}
		}
		return false;
	}

	public Hashtable getProperties() {
		return properties;
	}

	public void validate() throws Exception {
		if (cvsroot == null) {
			cvsroot = System.getProperty("CVSROOT");
		}
		if (cvsroot == null ) {
			throw new Exception("cvsroot' is a required attribute on CVS");
		}
	}

	/**
	 * Returns a List of Modifications detailing all the changes between the
	 * last build and the latest revision at the repository
	 * 
	 * @param fromDate
	 *            last build time
	 * @return maybe empty, never null.
	 */
	public RlogDocument getRlog(Date fromDate, Date toDate) {

		RlogDocument rlogDocument = RlogDocument.Factory.newInstance();
		rlogDocument.addNewRlog();
		
		
		try {
			cvsRLog(rlogDocument,createCVSRlogCommand(fromDate, toDate, this.getModules() ));
		} catch (Exception e) {
			e.printStackTrace();
			Trace.trace(Trace.SEVERE,
					"Log command failed to execute succesfully", e);
		}
		rlogDocument.getRlog().setStartdate(formatCVSDate(fromDate));
		rlogDocument.getRlog().setEnddate(formatCVSDate(toDate));
		rlogDocument.getRlog().addNewCounts();
		Delta[] darray = rlogDocument.getRlog().getCvsdeltaArray();
		int nDeltas = darray.length;
		int nRevs = 0;
		for (int i = 0; i < darray.length; i++) {
			Delta delta = darray[i];
			nRevs += delta.getRevisionArray().length;
			
		}
		rlogDocument.getRlog().getCounts().setNumDeltas(""+nDeltas);
		rlogDocument.getRlog().getCounts().setNumRevisions(""+nRevs);
		return rlogDocument;
	}



	/**
	 * @param lastBuildTime
	 * @param checkTime
	 * @return CommandLine for "cvs -d CVSROOT -q log -N -dlastbuildtime<checktime "
	 */
	public String[] createCVSRlogCommand(Date lastBuildTime, Date checkTime, String[] modules)
			throws Exception {

		ArrayList commands = new ArrayList();
		commands.add("cvs");
		commands.add("-d");
		commands.add(cvsroot);
		commands.add("-q");
		commands.add("rlog");
		boolean useHead = tag == null || tag.equals("HEAD") || tag.equals("");
		if(useHead)
			commands.add("-N");
		String dateRange = formatCVSDate(lastBuildTime) + " < "
		+ formatCVSDate(checkTime);
		commands.add("-d"+dateRange+"");
		if(useHead)
			commands.add("-b");
		else{
			commands.add("-r");
			commands.add(tag);
		}
		for (int i = 0; i < modules.length; i++) {
			commands.add(modules[i]);
		}
			

		return (String[]) commands.toArray(new String[commands.size()]);
	}

	public static String formatCVSDate(Date date) {
		return CVSDATE.format(date);
	}
	public static Date parseCVSDate(String date) throws ParseException {
		return CVSDATE.parse(date);
	}

	/**
	 * Parses the input stream, which should be from the cvs log command. This
	 * method will format the data found in the input stream into a List of
	 * Modification instances.
	 * 
	 * @param input
	 *            InputStream to get log data from.
	 * @return List of Modification elements, maybe empty never null.
	 * @throws IOException
	 */
	protected RlogDocument parseCVSDelta(RlogDocument rlogDocument, InputStream input) throws IOException {
		BufferedReader reader = new BufferedReader(new InputStreamReader(input));

		// Read to the first RCS file name. The first entry in the log
		// information will begin with this line. A CVS_FILE_DELIMITER is NOT
		// present. If no RCS file lines are found then there is nothing to do.
		String line = readToNotPast(reader, CVS_RCSFILE_LINE, null);
		ArrayList mods = new ArrayList();

		while (line != null) {
			// Parse the single file entry, which may include several
			// modifications.
			String workingFileName = line.substring(CVS_RCSFILE_LINE.length());
			int lastSlashIndex = workingFileName.lastIndexOf("/");
			String fileName, folderName = null;
			fileName = workingFileName.substring(lastSlashIndex + 1);
			if (lastSlashIndex != -1) {
				folderName = workingFileName.substring(0, lastSlashIndex);
			}
			
			line = readToNotPast(reader, CVS_HEAD, CVS_BRANCH);
			String head = "";
			if(line != null){
				head = line.substring(CVS_HEAD.length());
			}
			CVSDelta delta = new CVSDelta();
			delta.file = fileName;
			delta.folder = folderName;
			delta.branch = this.tag;
			delta.head = head;
			
			List returnList = parseRevisions(fileName, folderName,reader);
			delta.setRevisions(returnList);
			
			if(returnList.size() > 0 )
				delta.toXML(rlogDocument.getRlog().addNewCvsdelta());
			
			// Read to the next RCS file line. The CVS_FILE_DELIMITER may have
			// been consumed by the parseEntry method, so we cannot read to it.
			line = readToNotPast(reader, CVS_RCSFILE_LINE, null);
		}

		return rlogDocument;
	}

	private void getRidOfLeftoverData(InputStream stream) {
		BufferedReader reader = new BufferedReader(
				new InputStreamReader(stream), 1024);
		try {
			String s = reader.readLine();
			while (s != null) {
				s = reader.readLine();
			}
		} catch (IOException e) {
			// do nothing
		} finally {
			try {
				reader.close();
			} catch (IOException e) {
				// do nothing
			}
		}
	}

	private RlogDocument cvsRLog(RlogDocument rlogDocuments, String[] command) throws Exception {
		for (int i = 0; i < command.length; i++) {
			Trace.trace(Trace.FINEST, command[i] + " ");
			
		}
		Process p = Runtime.getRuntime().exec(command);

		logErrorStream(p);
		InputStream cvsLogStream = p.getInputStream();
		parseCVSDelta(rlogDocuments,cvsLogStream);

		getRidOfLeftoverData(cvsLogStream);
		p.waitFor();
		p.getInputStream().close();
		p.getOutputStream().close();
		p.getErrorStream().close();

		return rlogDocuments;
	}


	private void logErrorStream(Process p) {
	}

	// (PENDING) Extract CVSEntryParser class
	/**
	 * Parses a single file entry from the reader. This entry may contain zero
	 * or more revisions. This method may consume the next CVS_FILE_DELIMITER
	 * line from the reader, but no further.
	 * 
	 * @param reader
	 *            Reader to parse data from.
	 * @return modifications found in this entry; maybe empty, never null.
	 * @throws IOException
	 */
	private List parseRevisions(String fileName, String folderName,  BufferedReader reader) throws IOException {
		ArrayList mods = new ArrayList();

		String nextLine = "";

		// Read to the working file name line to get the filename. It is ASSUMED
		// that a line will exist with the working file name on it.
		//String workingFileLine = readToNotPast(reader, CVS_WORKINGFILE_LINE,
		//		null);
		String branchRevisionName = parseBranchRevisionName(reader, tag);

		boolean newCVSVersion = isCvsNewOutputFormat();
		while (nextLine != null && !nextLine.startsWith(CVS_FILE_DELIM)) {
			nextLine = readToNotPast(reader, "revision", CVS_FILE_DELIM);
			if (nextLine == null) {
				// No more revisions for this file.
				break;
			}

			nextLine.length();
			StringTokenizer tokens = new StringTokenizer(nextLine, " ");
			tokens.nextToken();
			String revision = tokens.nextToken();
			if (tag != null && !tag.equals(CVS_HEAD_TAG)) {
				String itsBranchRevisionName = revision.substring(0, revision
						.lastIndexOf('.'));
				if (!itsBranchRevisionName.equals(branchRevisionName)) {
					break;
				}
			}

			// Read to the revision date. It is ASSUMED that each revision
			// section will include this date information line.
			nextLine = readToNotPast(reader, CVS_REVISION_DATE, CVS_FILE_DELIM);
			if (nextLine == null) {
				break;
			}

			tokens = new StringTokenizer(nextLine, " \t\n\r\f;");
			// First token is the keyword for date, then the next two should be
			// the date and time stamps.
			tokens.nextToken();
			String dateStamp = tokens.nextToken();
			String timeStamp = tokens.nextToken();

			// skips the +0000 part of new format
			if (newCVSVersion) {
				tokens.nextToken();
			}
			// The next token should be the author keyword, then the author
			// name.
			tokens.nextToken();
			String authorName = tokens.nextToken();

			// The next token should be the state keyword, then the state name.
			tokens.nextToken();
			String stateKeyword = tokens.nextToken();

			// if no lines keyword then file is added
			boolean isAdded = !tokens.hasMoreTokens();

			// All the text from now to the next revision delimiter or working
			// file delimiter constitutes the messsage.
			String message = "";
			nextLine = reader.readLine();
			boolean multiLine = false;

			while (nextLine != null && !nextLine.startsWith(CVS_FILE_DELIM)
					&& !nextLine.startsWith(CVS_REVISION_DELIM)) {

				if (multiLine) {
					message += NEW_LINE;
				} else {
					multiLine = true;
				}
				message += nextLine;

				// Go to the next line.
				nextLine = reader.readLine();
			}

			CVSRevision nextModification = new CVSRevision();
			nextModification.rev = revision;


			CVSRevision.RevisedFile modfile = nextModification
					.createModifiedFile(fileName, folderName);
			modfile.revision = nextModification.rev;

			try {
				if (newCVSVersion) {
					nextModification.date = CVSDATE.parse(dateStamp
							+ " " + timeStamp + " GMT");
				} else {
					nextModification.date = LOGDATE.parse(dateStamp
							+ " " + timeStamp + " GMT");
				}
			} catch (ParseException pe) {
				Trace.trace(Trace.SEVERE,
						"Error parsing cvs log for date and time", pe);
				return null;
			}

			nextModification.author = authorName;

			nextModification.comment = (message != null ? message : "");

			if (stateKeyword.equalsIgnoreCase(CVS_REVISION_DEAD)
					&& message.indexOf("was initially added on branch") != -1) {
				//Trace.trace(Trace.FINEST,
				//		"skipping branch addition activity for "
				//				+ nextModification);
				// this prevents additions to a branch from showing up as action
				// "deleted" from head
				continue;
			}

			if (stateKeyword.equalsIgnoreCase(CVS_REVISION_DEAD)) {
				modfile.action = "deleted";
				if (propertyOnDelete != null) {
					properties.put(propertyOnDelete, "true");
				}
			} else if (isAdded) {
				modfile.action = "added";
			} else {
				modfile.action = "modified";
			}
			if (property != null) {
				properties.put(property, "true");
			}
			mods.add(nextModification);
		}
		return mods;
	}

	/**
	 * Find the CVS branch revision name, when the tag is not HEAD The reader
	 * will consume all lines up to the next description.
	 * 
	 * @param reader
	 *            the reader
	 * @param tag
	 *            may be null
	 * @return the branch revision name, or <code>null</code> if not
	 *         applicable or none was found.
	 * @throws IOException
	 */
	private static String parseBranchRevisionName(BufferedReader reader,
			final String tag) throws IOException {
		String branchRevisionName = null;

		if (tag != null && !tag.equals(CVS_HEAD_TAG)) {
			// Look for the revision of the form "tag: *.(0.)y ". this doesn't
			// work for HEAD
			// get line with branch revision on it.

			String branchRevisionLine = readToNotPast(reader,
					"\t" + tag + ": ", CVS_DESCRIPTION);

			if (branchRevisionLine != null) {
				// Look for the revision of the form "tag: *.(0.)y "
				branchRevisionName = branchRevisionLine
						.substring(tag.length() + 3);
				if (branchRevisionName.charAt(branchRevisionName
						.lastIndexOf(".") - 1) == '0') {
					branchRevisionName = branchRevisionName.substring(0,
							branchRevisionName.lastIndexOf(".") - 2)
							+ branchRevisionName.substring(branchRevisionName
									.lastIndexOf("."));
				}
			}
		}
		return branchRevisionName;
	}

	/**
	 * This method will consume lines from the reader up to the line that begins
	 * with the String specified but not past a line that begins with the
	 * notPast String. If the line that begins with the beginsWith String is
	 * found then it will be returned. Otherwise null is returned.
	 * 
	 * @param reader
	 *            Reader to read lines from.
	 * @param beginsWith
	 *            String to match to the beginning of a line.
	 * @param notPast
	 *            String which indicates that lines should stop being consumed,
	 *            even if the begins with match has not been found. Pass null to
	 *            this method to ignore this string.
	 * @return String that begin as indicated, or null if none matched to the
	 *         end of the reader or the notPast line was found.
	 * @throws IOException
	 */
	private static String readToNotPast(BufferedReader reader,
			String beginsWith, String notPast) throws IOException {
		boolean checkingNotPast = notPast != null;

		String nextLine = reader.readLine();
		while (nextLine != null && !nextLine.startsWith(beginsWith)) {
			if (checkingNotPast && nextLine.startsWith(notPast)) {
				return null;
			}
			nextLine = reader.readLine();
		}

		return nextLine;
	}

	public String[] getModules() {
		return modules;
	}
	

	public void setModules(String[] modules) {
		this.modules = modules;
	}
	

}
